﻿using System;
using System.ComponentModel;
using SlimDX;
using System.Windows.Media;

namespace SlimDXRenderSystem.Core
{
    public class Material
    {
        private static readonly Vector4 DEF_DIFFUSE = new Vector4(0.95f, 0.95f, 0.95f, 1.0f);
        private static readonly Vector4 DEF_SPECULAR = new Vector4(1.0f, 1.0f, 1.0f, 1.0f);
        private static readonly Vector4 DEF_EMISSIVE = new Vector4(0.0f, 0.0f, 0.0f, 1.0f);
        private static readonly Vector4 DEF_AMBIENT = new Vector4(0.05f, 0.05f, 0.05f, 1.0f);
        private const int DEF_SPECULAR_SHININESS = 50;

        private MaterialStruct material;

        [Browsable(false)]
        public MaterialStruct MaterialStruct
        {
            get { return material; }
            set { material = value; }
        }

        public Color Ambient
        {
            get { return Color.FromScRgb(material.Ambient.W, material.Ambient.X, material.Ambient.Y, material.Ambient.Z); }
            set { material.Ambient = new Vector4(value.ScR, value.ScG, value.ScB, 1); }
        }

        public Color Diffuse
        {
            get { return Color.FromScRgb(material.Diffuse.W, material.Diffuse.X, material.Diffuse.Y, material.Diffuse.Z); }
            set { material.Diffuse = new Vector4(value.ScR, value.ScG, value.ScB, 1); }
        }

        public Color Specular
        {
            get { return Color.FromScRgb(material.Specular.W, material.Specular.X, material.Specular.Y, material.Specular.Z); }
            set { material.Specular = new Vector4(value.ScR, value.ScG, value.ScB, 1); }
        }

        public Color Emissive
        {
            get { return Color.FromScRgb(material.Emissive.W, material.Emissive.X, material.Emissive.Y, material.Emissive.Z); }
            set { material.Emissive = new Vector4(value.ScR, value.ScG, value.ScB, 1); }
        }

        public float SpecularShininess
        {
            get { return material.Shininess; }
            set
            {
                if (value < 0)
                    material.Shininess = 0.0f;
                else
                    material.Shininess = value;
            }
        }

        public Material()
        {
            material = new MaterialStruct();
            material.Diffuse = DEF_DIFFUSE;
            material.Specular = DEF_SPECULAR;
            material.Shininess = DEF_SPECULAR_SHININESS;
            material.Emissive = DEF_EMISSIVE;
            material.Ambient = DEF_AMBIENT;
        }
    }
}
