﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using SlimDX;
using SlimDXRenderSystem.Managers;

namespace SlimDXRenderSystem.Core
{
    public class Movable
    {
        #region Fields
        private String name;

        private Vector3 vPosition;
        //private Vector3 vRotation;
        private Vector3 vScale;

        private Matrix matLocalWorld;
        private Matrix matTranslation;
        private Matrix matRotation;
        private Matrix matScale;
        

        #region Properties
        [Category("Information"), ReadOnly(true)]
        public String Name
        {
            get { return name; }
            set { name = value; }
        }

        [Category("Transformations")]
        public Vector3 Position
        {
            get { return vPosition; }
            set
            {
                vPosition = value;
                CalcMatrices();
            }
        }

        [Category("Transformations")]
        public Vector3 Scale
        {
            get { return vScale; }
            set
            {
                vScale = value;
                CalcMatrices();
            }
        }

        /*[Category("Transformations")]
        public Vector3 Rotation
        {
            get { return vRotation; }
            set
            {
                vRotation = value;
                CalcMatrices();
            }
        }*/

        [Browsable(false)]
        public Matrix LocalWorld
        {
            get { return matLocalWorld; }
            set { matLocalWorld = value; }
        }

        [Browsable(false)]
        public Matrix Translation
        {
            get { return matTranslation; }
        }

        [Browsable(false)]
        public Matrix RotationMatrix
        {
            get { return matRotation; }
            set { matRotation = value; }
        }

        [Browsable(false)]
        public Matrix ScaleMatrix
        {
            get { return matScale; }
        }
        #endregion
        #endregion

        #region Private Methods
        public void CalcMatrices()
        {
            matTranslation = Matrix.Translation(vPosition);
            matScale = Matrix.Scaling(vScale);
            //matRotation = Matrix.RotationYawPitchRoll(vRotation.X, vRotation.Y, vRotation.Z);            
            matLocalWorld = matRotation * matScale * matTranslation;
        }

        #endregion

        #region Public Methods
        public Movable()
        {
            this.name = "untitled";

            vPosition = Vector3.Zero;
            //vRotation = Vector3.Zero;
            vScale = new Vector3(1.0f, 1.0f, 1.0f);

            matTranslation = Matrix.Identity;
            matScale = Matrix.Scaling(vScale);
            matRotation = Matrix.Identity;
            matLocalWorld = Matrix.Identity;
        }

        public void Move(Vector3 position)
        {
            Move(position.X, position.Y, position.Z);
        }

        public void Move(float x, float y, float z)
        {
            this.vPosition.X += x;
            this.vPosition.Y += y;
            this.vPosition.Z += z;
            CalcMatrices();
        }

        public void MoveTo(Vector3 position)
        {
            MoveTo(position.X, position.Y, position.Z);
        }

        public void MoveTo(float x, float y, float z)
        {
            this.vPosition.X = x;
            this.vPosition.Y = y;
            this.vPosition.Z = z;
            CalcMatrices();
        }

        /*public void Rotate(Vector3 rotation)
        {
            Rotate(rotation.X, rotation.Y, rotation.Z);
        }*/

        /*public void Rotate(float rx, float ry, float rz)
        {
            vRotation.X += rx;
            vRotation.Y += ry;
            vRotation.Z += rz;
            CalcMatrices();
        }*/

        public void Rotate(Vector3 axis, float angle)
        {
            Matrix r = Matrix.RotationAxis(axis, angle);
            this.matRotation = this.matRotation * r;
            CalcMatrices();
        }

        public void Scaling(Vector3 scale)
        {
            Scaling(scale.X, scale.Y, scale.Z);
        }

        public void Scaling(float sx, float sy, float sz)
        {
            vScale.X += sx;
            vScale.Y += sy;
            vScale.Z += sz;
            CalcMatrices();
        }

        public override string ToString()
        {
            return this.Name;
        }
        #endregion
    }
}
