﻿using System;

namespace SlimDXRenderSystem.Core
{
    /// <summary>
    /// Thread-safe Singleton Pattern
    /// </summary>
    /// <typeparam name="T">Type witch will be used for singleton.</typeparam>
    public class Singleton<T> where T : class, new()
    {
        static Singleton()
        {
        }

        public static T Instance
        {
            get
            {
                return SingletonCreator.instance;
            }
        }

        class SingletonCreator
        {
            static SingletonCreator()
            {
            }

            internal static readonly T instance = new T();
        }
    }
}
