﻿using System;
using SlimDX;
using SlimDXRenderSystem.Direct3DBase;

namespace SlimDXRenderSystem.Managers
{
    public class EffectManager : Core.Singleton<EffectManager>
    {
        private Effect superEffect;

        public Effect SuperEffect
        {
            get { return superEffect; }
            set { superEffect = value; }
        }

        public void LoadEffectFromFile(string filename)
        {
            switch(RenderManager.Instance.D3dVersion)
            {
                case D3DVersion.Direct3D9:
                    superEffect = new Direct3D9.Effect9();
                    break;
            }
            superEffect.LoadFromFile(filename);
        }

        public void LoadEffectFromResource(string filename)
        {
            switch (RenderManager.Instance.D3dVersion)
            {
                case D3DVersion.Direct3D9:
                    superEffect = new Direct3D9.Effect9();
                    break;
            }
            superEffect.LoadFromResource(filename);
            SetupAllParams();
        }

        public void SetParameter<T>(string param, T value) where T : struct
        {
            superEffect.SetParameter<T>(param, value);
        }

        public void SetParameter(string param, Color3 value)
        {
            SetParameter<Vector3>(param, new Vector3(value.Red, value.Green, value.Blue));
        }

        public void SetParameter(string param, Color4 value)
        {
            SetParameter<Vector4>(param, new Vector4(value.Red, value.Green, value.Blue, value.Alpha));
        }

        public void SetParameter(string param, System.Drawing.Color value)
        {
            Vector4 color = new Vector4((float)value.R / 255, (float)value.G / 255, (float)value.B / 255, (float)value.A / 255);
            SetParameter<Vector4>(param, color);
        }

        public T GetParameter<T>(string param) where T : struct
        {
            return superEffect.GetParameter<T>(param);
        }

        public void SetupAllParams()
        {
            try
            {
                SetParameter(Effect.AMBIENTCOLOR, RenderManager.Instance.AmbientLight);
                SetParameter(Effect.AMBIENTINTENSITY, RenderManager.Instance.AmbientIntensity);
                SetParameter(Effect.LIGHTING, RenderManager.Instance.Lighting);
            }
            catch (Exception e)
            {
                throw e;
            }
        }

        public void Dispose()
        {
            if (superEffect != null && !superEffect.Disposed)
            {
                superEffect.Dispose();
                superEffect = null;
            }
        }
    }
}
