﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Framework;


namespace MeshBrowser.SlimDXRenderer
{
    public class GeoidSource
    {
        int depth = 2;

        public int Depth
        {
            get { return depth; }
            set { depth = value; }
        }        

        public TriangleMesh result = null;

        public void Execute(double scale, Point3D center)
        {
            
                List<int> triangles = new List<int>();
                List<double> vertices = new List<double>();

                /*triangles = initTetraTris();
                vertices = initTetraVerts();*/
                triangles = initIcosaTris();
                vertices = initIcosaVerts();

                for (int i = 0; i < depth; i++)
                {
                    splitTrianlges(triangles, vertices);
                }

                result = new TriangleMesh();

                Point3D[] points = new Point3D[vertices.Count / 3];
                Point3D[] normals = new Point3D[points.Length];
                for (int i = 0; i < points.Length; i++)
                {
                    points[i] = new Point3D(vertices[3 * i], vertices[3 * i + 1], vertices[3 * i + 2]);
                    normals[i] = points[i];
                    points[i] *= scale;
                    points[i] += center;
                }                               

                Triangle[] tris = new Triangle[triangles.Count / 3];
                for (int i = 0; i < tris.Length; i++)
                {
                    tris[i] = new Triangle(triangles[3 * i], triangles[3 * i + 1], triangles[3 * i + 2]);
                }

                result.Points = points;
                result.Triangles = tris;
                result.Normals = normals;                
        }

        private static void splitTrianlges(List<int> triangles, List<double> vertices)
        {
            Dictionary<Line, int> nv = new Dictionary<Line, int>();
            List<int> newTriangles = new List<int>();
            List<double> newVertices = new List<double>(vertices);

            for (int i = 0; i < (triangles.Count / 3); i++)
            {
                int i1 = newVertices.Count / 3;
                Line line = new Line(triangles[3 * i], triangles[3 * i + 1], false);
                if (nv.ContainsKey(line))
                {
                    i1 = nv[line];
                }
                else
                {
                    double x = (vertices[triangles[3 * i] * 3] + vertices[triangles[3 * i + 1] * 3]) / 2;
                    double y = (vertices[triangles[3 * i] * 3 + 1] + vertices[triangles[3 * i + 1] * 3 + 1]) / 2;
                    double z = (vertices[triangles[3 * i] * 3 + 2] + vertices[triangles[3 * i + 1] * 3 + 2]) / 2;
                    double l = Math.Sqrt(x * x + y * y + z * z);
                    //l = 1;
                    newVertices.Add(x / l);
                    newVertices.Add(y / l);
                    newVertices.Add(z / l);
                    nv.Add(line, i1);
                }

                int i2 = newVertices.Count / 3;
                line = new Line(triangles[3 * i], triangles[3 * i + 2], false);
                if (nv.ContainsKey(line))
                {
                    i2 = nv[line];
                }
                else
                {
                    double x = (vertices[triangles[3 * i] * 3] + vertices[triangles[3 * i + 2] * 3]) / 2;
                    double y = (vertices[triangles[3 * i] * 3 + 1] + vertices[triangles[3 * i + 2] * 3 + 1]) / 2;
                    double z = (vertices[triangles[3 * i] * 3 + 2] + vertices[triangles[3 * i + 2] * 3 + 2]) / 2;
                    double l = Math.Sqrt(x * x + y * y + z * z);
                    //l = 1;
                    newVertices.Add(x / l);
                    newVertices.Add(y / l);
                    newVertices.Add(z / l);
                    nv.Add(line, i2);
                }

                int i3 = newVertices.Count / 3;
                line = new Line(triangles[3 * i + 1], triangles[3 * i + 2], false);
                if (nv.ContainsKey(line))
                {
                    i3 = nv[line];
                }
                else
                {
                    double x = (vertices[triangles[3 * i + 1] * 3] + vertices[triangles[3 * i + 2] * 3]) / 2;
                    double y = (vertices[triangles[3 * i + 1] * 3 + 1] + vertices[triangles[3 * i + 2] * 3 + 1]) / 2;
                    double z = (vertices[triangles[3 * i + 1] * 3 + 2] + vertices[triangles[3 * i + 2] * 3 + 2]) / 2;
                    double l = Math.Sqrt(x * x + y * y + z * z);
                    //l = 1;
                    newVertices.Add(x / l);
                    newVertices.Add(y / l);
                    newVertices.Add(z / l);
                    nv.Add(line, i3);
                }

                newTriangles.Add(i1);
                newTriangles.Add(i3);
                newTriangles.Add(i2);

                newTriangles.Add(i1);
                newTriangles.Add(i2);
                newTriangles.Add(triangles[3 * i]);

                newTriangles.Add(i3);
                newTriangles.Add(i1);
                newTriangles.Add(triangles[3 * i + 1]);

                newTriangles.Add(i2);
                newTriangles.Add(i3);
                newTriangles.Add(triangles[3 * i + 2]);
            }

            triangles.Clear();
            triangles.AddRange(newTriangles);
            vertices.Clear();
            vertices.AddRange(newVertices);

        }

        private static List<double> initIcosaVerts()
        {
            double[] vertices = new double[12 * 3];
            double phiaa = 26.56505; /* phi needed for generation */
            double r = 1.0; /* any radius in which the polyhedron is inscribed */
            double phia = Math.PI * phiaa / 180.0; /* 2 sets of four points */
            double theb = Math.PI * 36.0 / 180.0;  /* offset second set 36 degrees */
            double the72 = Math.PI * 72.0 / 180;   /* step 72 degrees */
            vertices[0] = 0.0;
            vertices[1] = 0.0;
            vertices[2] = r;
            vertices[11*3] = 0.0;
            vertices[11*3+1] = 0.0;
            vertices[11*3+2] = -r;
            double the = 0.0;
            for (int i = 1; i < 6; i++)
            {
                vertices[i * 3] = r * Math.Cos(the) * Math.Cos(phia);
                vertices[i * 3 + 1] = r * Math.Sin(the) * Math.Cos(phia);
                vertices[i * 3 + 2] = r * Math.Sin(phia);
                the = the + the72;
            }
            the = theb;
            for (int i = 6; i < 11; i++)
            {
                vertices[i * 3] = r * Math.Cos(the) * Math.Cos(-phia);
                vertices[i * 3 + 1] = r * Math.Sin(the) * Math.Cos(-phia);
                vertices[i * 3 + 2] = r * Math.Sin(-phia);
                the = the + the72;
            }
            return(new List<double>(vertices));
        }

        private static List<int> initIcosaTris()
        {            
            int[] result = new int[] {
                0, 1, 2, 
                0, 2, 3, 
                0, 3, 4, 
                0, 4, 5, 
                0, 5, 1, 
                11, 7, 6, 
                11, 8, 7, 
                11, 9, 8, 
                11, 10, 9, 
                11, 6, 10, 
                1, 6, 2, 
                2, 7, 3, 
                3, 8, 4, 
                4, 9, 5, 
                5, 10, 1, 
                6, 7, 2, 
                7, 8, 3, 
                8, 9, 4, 
                9, 10, 5, 
                10, 6, 1};
            return (new List<int>(result));
        }

        private static List<double> initTetraVerts()
        {
            List<double> result = new List<double>();
            double y = -Math.Sin(Math.PI / 6);
            double z = Math.Cos(Math.PI / 6);

            result.Add(-Math.Sin(Math.PI * 2.0 / 3.0) * z);
            result.Add(y);
            result.Add(Math.Cos(Math.PI * 2.0 / 3.0) * z);

            result.Add(Math.Sin(Math.PI * 2.0 / 3.0) * z);
            result.Add(y);
            result.Add(Math.Cos(Math.PI * 2.0 / 3.0) * z);

            result.Add(0);
            result.Add(y);
            result.Add(z);

            result.Add(0);
            result.Add(1);
            result.Add(0);

            return (result);
        }

        private static List<int> initTetraTris()
        {
            List<int> result = new List<int>();

            result.Add(0);
            result.Add(1);
            result.Add(2);

            result.Add(0);
            result.Add(2);
            result.Add(3);

            result.Add(2);
            result.Add(1);
            result.Add(3);

            result.Add(1);
            result.Add(0);
            result.Add(3);

            return (result);
        }
    }
}
