﻿using System;
using SlimDXRenderSystem;
using SlimDX.DirectInput;
using SlimDX;
using SceneMgr = SlimDXRenderSystem.Managers.SceneManager;
using InputMgr = SlimDXRenderSystem.Managers.InputManager;
using RenderMgr = SlimDXRenderSystem.Managers.RenderManager;

namespace MeshBrowser.SlimDXRenderer
{
    public class InputSetting
    {
        public InputSetting()
        {
            InputMgr.Instance.AddMouseClickAction(MouseObject.Button1, MousePick);
            InputMgr.Instance.AddMouseMoveButAction(MouseObject.Button1, RotateCamera);
            InputMgr.Instance.AddMouseMoveButAction(MouseObject.Button2, MoveObject);
            InputMgr.Instance.AddMouseMoveButAction(MouseObject.Button3, MoveCamera);
            InputMgr.Instance.AddMouseMoveKeyAction(Key.L, RotateLight);
            InputMgr.Instance.AddMouseMoveAction(ZoomCamera);

            InputMgr.Instance.AddKeyClickAction(Key.Comma, PreviousObject);
            InputMgr.Instance.AddKeyClickAction(Key.Period, NextObject);
            InputMgr.Instance.AddKeyClickAction(Key.G, CursorPick);

            InputMgr.Instance.AddKeyIsPressedAction(Key.W, MoveCameraForward);
            InputMgr.Instance.AddKeyIsPressedAction(Key.S, MoveCameraBackward);
            InputMgr.Instance.AddKeyIsPressedAction(Key.A, MoveCameraLeft);
            InputMgr.Instance.AddKeyIsPressedAction(Key.D, MoveCameraRight);
            InputMgr.Instance.AddKeyIsPressedAction(Key.E, MoveCameraUp);
            InputMgr.Instance.AddKeyIsPressedAction(Key.Q, MoveCameraDown);
            InputMgr.Instance.AddKeyIsPressedAction(Key.R, CameraZoomIn);
            InputMgr.Instance.AddKeyIsPressedAction(Key.T, CameraZoomOut);            

            InputMgr.Instance.AddControllerMoveKeyAction(Key.K, RotateObjectController);
            InputMgr.Instance.AddControllerMoveKeyAction(Key.J, RotateLightController);
            InputMgr.Instance.AddControllerMoveKeyAction(Key.H, MoveCursor);
            InputMgr.Instance.AddControllerMoveKeyAction(Key.F, MoveObjectController);
            //InputMgr.Instance.AddControllerMoveKeyAction(Key.F, MoveCameraController);

            InputMgr.Instance.AddControllerRollKeyAction(Key.K, RollCameraController);

            
        }

        public void MousePick()
        {
            if (ModuleManager.Instance.Pick != PickType.NONE)
                ModuleManager.Instance.Picking();
        }

        public void CursorPick()
        {
            if (ModuleManager.Instance.Pick != PickType.NONE)
                ModuleManager.Instance.CursorPicking();
        }
        
        public void RotateCamera(int mousex, int mousey, int mousez)
        {
            SceneMgr.Instance.Camera.Rotate((float)mousex / 10, (float)mousey / 10, 0.0f);
        }

        public void RotateCameraController(float x, float y, float z)
        {
            float m = ModuleManager.Instance.Rendersetting.Sensitivity / 100.0f;
            SceneMgr.Instance.Camera.Rotate(m*x/2 , -m*y/2 , 0.0f);
        }

        public void RotateObjectController(float x, float y, float z)
        {
            float m = ModuleManager.Instance.Rendersetting.Sensitivity / 100.0f;
            ModuleManager.Instance.MeshGroup.Rotate(SceneMgr.Instance.Camera.YAxis, -m*x / 100f);
            ModuleManager.Instance.MeshGroup.Rotate(SceneMgr.Instance.Camera.XAxis, m*y / 100f);
        }

        public void RollCameraController(float rollDiff, float heightDiff)
        {
            float m = ModuleManager.Instance.Rendersetting.Sensitivity / 100.0f;
            //SceneMgr.Instance.Camera.Rotate(0.0f, 0.0f, m*57.29f*rollDiff);
            ModuleManager.Instance.MeshGroup.Rotate(SceneMgr.Instance.Camera.ZAxis, m * rollDiff);
            SceneMgr.Instance.Camera.Zoom(-m*heightDiff/10.0f);
        }

        public void MoveCamera(int mousex, int mousey, int mousez)
        {
            if (InputMgr.Instance.Keystate.IsPressed(Key.LeftShift))
                SceneMgr.Instance.Camera.Move((float)mousex / 10, 0.0f, (float)mousey / 10);
            else
                SceneMgr.Instance.Camera.Move((float)mousex / 10, (float)mousey / 10, 0.0f);
        }

        public void ZoomCamera(int mousex, int mousey, int mousez)
        {

            if (InputMgr.Instance.Keystate.IsPressed(Key.LeftShift))
            {
                SceneMgr.Instance.Camera.Zoom((float)-mousez / 25);
            }
            else
            {
                SceneMgr.Instance.Camera.Zoom((float)-mousez / 100);
            }
        }

        public void MoveObject(int mousex, int mousey, int mousez)
        {
            float x = (float)mousex / 100;
            float y = (float)mousey / 100;
            Vector3 move = Vector3.Zero;
            if (InputMgr.Instance.Keystate.IsPressed(Key.LeftShift))
            {
                move = SceneMgr.Instance.Camera.XAxis * x;
                move += SceneMgr.Instance.Camera.ZAxis * y;
            }
            else if (InputMgr.Instance.Keystate.IsPressed(Key.LeftControl))
            {
                if (InputMgr.Instance.Keystate.IsPressed(Key.LeftAlt))
                {
                    ModuleManager.Instance.MeshGroup.Scaling(y, y, y);
                }
                move = new Vector3(x, -y, 0.0f);
            }
            else if (InputMgr.Instance.Keystate.IsPressed(Key.LeftAlt))
            {
                //ModuleManager.Instance.MeshGroup.Rotate(-x, -y, 0.0f);
                ModuleManager.Instance.MeshGroup.Rotate(SceneMgr.Instance.Camera.YAxis, -x/5f);
                ModuleManager.Instance.MeshGroup.Rotate(SceneMgr.Instance.Camera.XAxis, -y/5f);
            }
            else
            {
                move = SceneMgr.Instance.Camera.XAxis * x;
                move -= SceneMgr.Instance.Camera.YAxis * y;
            }

            ModuleManager.Instance.MeshGroup.Move(move);
        }

        public void MoveObjectController(float mousex, float mousey, float mousez)
        {
            Vector3 move = Vector3.Zero;
            move = SceneMgr.Instance.Camera.XAxis * mousex/60;
            move += SceneMgr.Instance.Camera.YAxis * mousey/60;
            move -= SceneMgr.Instance.Camera.ZAxis * mousez/60;

            float m = ModuleManager.Instance.Rendersetting.Sensitivity / 100.0f;
            ModuleManager.Instance.MeshGroup.Move(move*m);
        }

        public void MoveCameraController(float mousex, float mousey, float mousez)
        {
            float m = ModuleManager.Instance.Rendersetting.Sensitivity / 100.0f;
            SceneMgr.Instance.Camera.Move(m*mousex/3, -m*mousey/3, -m*mousez/3);
        }

        public void MoveCursor(float mousex, float mousey, float mousez)
        {
            float x = (float)mousex / 100;
            float y = (float)mousey / 100;
            float z = (float)mousez / 100;
            Vector3 move = Vector3.Zero;
            /*if (InputMgr.Instance.Keystate.IsPressed(Key.LeftShift))
            {
                move = SceneMgr.Instance.Camera.XAxis * x;
                move += SceneMgr.Instance.Camera.ZAxis * y;
            }
            else if (InputMgr.Instance.Keystate.IsPressed(Key.LeftControl))
            {
                if (InputMgr.Instance.Keystate.IsPressed(Key.LeftAlt))
                {
                    ModuleManager.Instance.MeshGroup.Scaling(y, y, y);
                }
                move = new Vector3(x, -y, 0.0f);
            }
            else if (InputMgr.Instance.Keystate.IsPressed(Key.LeftAlt))
            {
                ModuleManager.Instance.MeshGroup.Rotate(-x, -y, 0.0f);
            }
            else*/
            {
                move = SceneMgr.Instance.Camera.XAxis * x;
                move += SceneMgr.Instance.Camera.YAxis * y;
                move -= SceneMgr.Instance.Camera.ZAxis * z;
            }

            float m = ModuleManager.Instance.Rendersetting.Sensitivity / 100.0f;
            ModuleManager.Instance.CursorGroup.Move(move*m);
            SceneMgr.Instance.CursorLight.Position += move*m;

            CursorPick();
        }

        public void PreviousObject()
        {
            ModuleManager.Instance.CurrentMesh -= 1;
        }

        public void NextObject()
        {
            ModuleManager.Instance.CurrentMesh += 1;
        }

        
        public void RotateLight(int mousex, int mousey, int mousez)
        {
            float lightElev = (float)Math.Asin(SceneMgr.Instance.Light.Direction.Y);
            float lightAzim = (float)Math.Atan2(SceneMgr.Instance.Light.Direction.Z, SceneMgr.Instance.Light.Direction.X);

            lightElev += (float)mousey / 100;
            lightAzim += (float)mousex / 100;
            SceneMgr.Instance.Light.Direction = new Vector3((float)(Math.Cos(lightAzim) * Math.Cos(lightElev)),
                                                    (float)(Math.Sin(lightElev)),
                                                    (float)(Math.Sin(lightAzim) * Math.Cos(lightElev)));
        }

        public void RotateLightController(float mousex, float mousey, float mousez)
        {
            float lightElev = (float)Math.Asin(SceneMgr.Instance.Light.Direction.Y);
            float lightAzim = (float)Math.Atan2(SceneMgr.Instance.Light.Direction.Z, SceneMgr.Instance.Light.Direction.X);

            float m = ModuleManager.Instance.Rendersetting.Sensitivity / 100.0f;
            lightElev -= m*mousey / 100;
            lightAzim += m*mousex / 100;
            SceneMgr.Instance.Light.Direction = new Vector3((float)(Math.Cos(lightAzim) * Math.Cos(lightElev)),
                                                    (float)(Math.Sin(lightElev)),
                                                    (float)(Math.Sin(lightAzim) * Math.Cos(lightElev)));
        }

        public void MoveCameraForward(double time)
        {
            SceneMgr.Instance.Camera.Move(0.0f, 0.0f, -(float)(time / 10));
        }

        public void MoveCameraBackward(double time)
        {
            SceneMgr.Instance.Camera.Move(0.0f, 0.0f, (float)(time/10));
        }

        public void MoveCameraLeft(double time)
        {
            SceneMgr.Instance.Camera.Move((float)(time/10), 0.0f, 0.0f);
        }

        public void MoveCameraRight(double time)
        {
            SceneMgr.Instance.Camera.Move(-(float)(time / 10), 0.0f, 0.0f);
        }

        public void MoveCameraUp(double time)
        {
            SceneMgr.Instance.Camera.Move(0.0f, (float)(time / 10), 0.0f);
        }

        public void MoveCameraDown(double time)
        {
            SceneMgr.Instance.Camera.Move(0.0f, -(float)(time / 10), 0.0f);
        }

        public void CameraZoomIn(double time)
        {
            SceneMgr.Instance.Camera.Zoom(-(float)(time / 100));
        }

        public void CameraZoomOut(double time)
        {
            SceneMgr.Instance.Camera.Zoom((float)(time / 100));
        }
    }
}
