﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Framework;

namespace MeshBrowser.SlimDXRenderer
{
    public class KDTree
    {
        class Node
        {
            public Node left, right;
            public int point;
            public int axis;            
        }

        Node root;

        Point3D[] pnts;

        Point3D minPoint, maxPoint;

        public KDTree(Point3D[] points)
        {
            this.pnts = points;
            List<int> list = new List<int>();
            for (int i = 0; i < points.Length; i++)
                list.Add(i);            
            root = new Node();
            addChildren(root, list, 0, 0);

            maxPoint = new Point3D(double.NegativeInfinity, double.NegativeInfinity, double.NegativeInfinity);
            minPoint = new Point3D(double.PositiveInfinity, double.PositiveInfinity, double.PositiveInfinity);
            for (int i = 0; i < pnts.Length; i++)
            {
                if (pnts[i].X < minPoint.X)
                    minPoint.X = pnts[i].X;
                if (pnts[i].X > maxPoint.X)
                    maxPoint.X = pnts[i].X;
                if (pnts[i].Y < minPoint.Y)
                    minPoint.Y = pnts[i].Y;
                if (pnts[i].Y > maxPoint.Y)
                    maxPoint.Y = pnts[i].Y;
                if (pnts[i].Z < minPoint.Z)
                    minPoint.Z = pnts[i].Z;
                if (pnts[i].Z > maxPoint.Z)
                    maxPoint.Z = pnts[i].Z;
            }
        }

        double smDist, smDistSq;
        int nearest;

        public void findNearest(Point3D p, List<double> dist, List<int> indices, double maxdist)
        {
            indices.Clear();
            dist.Clear();
            checkSubtree(root, indices, dist, maxdist*maxdist, maxdist, p);
        }

        public int findNearestToPlane(Point3D n, double d, Point3D a, Point3D b, Point3D c, double minDist)
        {
            this.A = a;
            this.B = b;
            this.C = c;
            this.minDist = minDist;
            smDist = double.MaxValue;
            nearest = -1;
            checkSubTreePlane(n, d, minPoint, maxPoint, root);

            //Console.WriteLine("{0} checked, {1} skipped", checkCount, skipCount);

            return (nearest);
        }

        int checkCount = 0;
        int skipCount = 0;

        private void checkSubTreePlane(Point3D n, double d, Point3D min, Point3D max, Node node)
        {
            if (node == null)
                return;

            // compute critical points
            Point3D cp1 = new Point3D();
            Point3D cp2 = new Point3D();

            if (n.X > 0)
            {
                cp1.X = max.X;
                cp2.X = min.X;
            }
            else
            {
                cp1.X = min.X;
                cp2.X = max.X;
            }

            if (n.Y > 0)
            {
                cp1.Y = max.Y;
                cp2.Y = min.Y;
            }
            else
            {
                cp1.Y = min.Y;
                cp2.Y = max.Y;
            }

            if (n.Z > 0)
            {
                cp1.Z = max.Z;
                cp2.Z = min.Z;
            }
            else
            {
                cp1.Z = min.Z;
                cp2.Z = max.Z;
            }

            // distance at critical points
            double cpv1 = n.DotProduct(cp1) + d;
            double cpv2 = n.DotProduct(cp2) + d;

            if ((cpv1 * cpv2 < 0) || (Math.Abs(cpv1) < smDist) || (Math.Abs(cpv2) < smDist))
            {
                // cell is intersected by the plane or near it, we must check it and proceed
                double p = Math.Abs(n.DotProduct(pnts[node.point]) + d);
                if (p < smDist)
                {
                    if (checkPoint(pnts[node.point]))
                    {
                        smDist = p;
                        nearest = node.point;
                    }
                }
                Point3D newMax = max;
                newMax[node.axis] = pnts[node.point][node.axis];
                checkSubTreePlane(n, d, min, newMax, node.left);

                Point3D newMin = min;
                newMin[node.axis] = pnts[node.point][node.axis];
                checkSubTreePlane(n, d, newMin, max, node.right);

                checkCount++;
            }
            else
                skipCount++;

        }

        Point3D A, B, C;

        double minDist;

        bool checkPoint(Point3D p)
        {
            //Point3D d = p - A;
            if (p.SquareDistanceTo(A) < minDist)
                return (false);
            //d = p - B;
            if (p.SquareDistanceTo(B) < minDist)
                return (false);
            //d = p - C;
            if (p.SquareDistanceTo(C) < minDist)
                return (false);
            return (true);
        }

        private void checkSubtree(Node n, List<int> indices, List<double> dist, double maxdistsq, double maxdist, Point3D p)
        {
            if (n == null)
                return;
            //Point3D v = p-pnts[n.point];
            //double d = v.DotProduct(v);
            double d = p.SquareDistanceTo(pnts[n.point]);
            double v = p[n.axis] - pnts[n.point][n.axis];
            if (d < maxdistsq)
            {
                indices.Add(n.point);
                dist.Add(Math.Sqrt(d));
            }
            if (v>0)
            {
                checkSubtree(n.right, indices, dist, maxdistsq, maxdist, p);
                if (v<maxdist)
                    checkSubtree(n.left, indices, dist, maxdistsq, maxdist, p);
            }
            else
            {
                checkSubtree(n.left, indices, dist, maxdistsq, maxdist, p);
                if ((-v) < maxdist)
                    checkSubtree(n.right, indices, dist, maxdistsq, maxdist, p);
            }
        }

        public int findNearest(Point3D p, out double dist)
        {
            smDist = double.MaxValue;
            smDistSq = double.MaxValue;
            nearest = -1;
            //searchSubtree(root, p);
            searchSubtreeX(root, p);
            dist = smDist;
            return (nearest);
        }

        void searchSubtree(Node n, Point3D p)
        {   
            // node point
            Point3D np = pnts[n.point];     
       
            if (p[n.axis] < np[n.axis])
            {
                if ((np[n.axis] - p[n.axis])<smDist)
                {
                    //Point3D dif = p - np;
                    //Point3D dif = p.Subtract(np);
                    //double dist = dif.DotProduct(dif);
                    double dist = p.SquareDistanceTo(np);
                    if (dist < smDistSq)
                    {
                        smDistSq = dist;
                        smDist = Math.Sqrt(dist);
                        nearest = n.point;
                    }
                }
                if (n.left != null)                
                {
                    searchSubtree(n.left, p);
                    if (n.right != null)
                    {
                        double minDist = np[n.axis] - p[n.axis];
                        if (minDist < smDist)
                            searchSubtree(n.right, p);
                    }
                }
            }
            else
            {
                if ((p[n.axis]-np[n.axis]) < smDist)
                {
                    //Point3D dif = p.Subtract(np);
                    //double dist = dif.DotProduct(dif);
                    double dist = p.SquareDistanceTo(np);
                    if (dist < smDistSq)
                    {
                        smDist = Math.Sqrt(dist);
                        smDistSq = dist;
                        nearest = n.point;
                    }
                }
                if (n.right != null)
                {
                    searchSubtree(n.right, p);
                    if (n.left != null)
                    {
                        double minDist = p[n.axis] - np[n.axis];
                        if (minDist < smDist)
                            searchSubtree(n.left, p);
                    }
                }
            }         
        }

        void searchSubtreeX(Node n, Point3D p)
        {
            // node point
            Point3D np = pnts[n.point];

            if (p.X < np.X)
            {
                if ((np.X - p.X) < smDist)
                {
                    //Point3D dif = p - np;
                    //Point3D dif = p.Subtract(np);
                    //double dist = dif.DotProduct(dif);
                    double dist = p.SquareDistanceTo(np);
                    if (dist < smDistSq)
                    {
                        smDistSq = dist;
                        smDist = Math.Sqrt(dist);
                        nearest = n.point;
                    }
                }
                if (n.left != null)
                {
                    searchSubtreeY(n.left, p);
                    if (n.right != null)
                    {
                        double minDist = np.X - p.X;
                        if (minDist < smDist)
                            searchSubtreeY(n.right, p);
                    }
                }
            }
            else
            {
                if ((p.X - np.X) < smDist)
                {
                    //Point3D dif = p.Subtract(np);
                    //double dist = dif.DotProduct(dif);
                    double dist = p.SquareDistanceTo(np);
                    if (dist < smDistSq)
                    {
                        smDist = Math.Sqrt(dist);
                        smDistSq = dist;
                        nearest = n.point;
                    }
                }
                if (n.right != null)
                {
                    searchSubtreeY(n.right, p);
                    if (n.left != null)
                    {
                        double minDist = p.X - np.X;
                        if (minDist < smDist)
                            searchSubtreeY(n.left, p);
                    }
                }
            }
        }

        void searchSubtreeY(Node n, Point3D p)
        {
            // node point
            Point3D np = pnts[n.point];

            if (p.Y < np.Y)
            {
                if ((np.Y - p.Y) < smDist)
                {
                    //Point3D dif = p - np;
                    //Point3D dif = p.Subtract(np);
                    //double dist = dif.DotProduct(dif);
                    double dist = p.SquareDistanceTo(np);
                    if (dist < smDistSq)
                    {
                        smDistSq = dist;
                        smDist = Math.Sqrt(dist);
                        nearest = n.point;
                    }
                }
                if (n.left != null)
                {
                    searchSubtreeZ(n.left, p);
                    if (n.right != null)
                    {
                        double minDist = np.Y - p.Y;
                        if (minDist < smDist)
                            searchSubtreeZ(n.right, p);
                    }
                }
            }
            else
            {
                if ((p.Y - np.Y) < smDist)
                {
                    //Point3D dif = p.Subtract(np);
                    //double dist = dif.DotProduct(dif);
                    double dist = p.SquareDistanceTo(np);
                    if (dist < smDistSq)
                    {
                        smDist = Math.Sqrt(dist);
                        smDistSq = dist;
                        nearest = n.point;
                    }
                }
                if (n.right != null)
                {
                    searchSubtreeZ(n.right, p);
                    if (n.left != null)
                    {
                        double minDist = p.Y - np.Y;
                        if (minDist < smDist)
                            searchSubtreeZ(n.left, p);
                    }
                }
            }
        }

        void searchSubtreeZ(Node n, Point3D p)
        {
            // node point
            Point3D np = pnts[n.point];

            if (p.Z < np.Z)
            {
                if ((np.Z - p.Z) < smDist)
                {
                    //Point3D dif = p - np;
                    //Point3D dif = p.Subtract(np);
                    //double dist = dif.DotProduct(dif);
                    double dist = p.SquareDistanceTo(np);
                    if (dist < smDistSq)
                    {
                        smDistSq = dist;
                        smDist = Math.Sqrt(dist);
                        nearest = n.point;
                    }
                }
                if (n.left != null)
                {
                    searchSubtreeX(n.left, p);
                    if (n.right != null)
                    {
                        double minDist = np.Z - p.Z;
                        if (minDist < smDist)
                            searchSubtreeX(n.right, p);
                    }
                }
            }
            else
            {
                if ((p.Z - np.Z) < smDist)
                {
                    //Point3D dif = p.Subtract(np);
                    //double dist = dif.DotProduct(dif);
                    double dist = p.SquareDistanceTo(np);
                    if (dist < smDistSq)
                    {
                        smDist = Math.Sqrt(dist);
                        smDistSq = dist;
                        nearest = n.point;
                    }
                }
                if (n.right != null)
                {
                    searchSubtreeX(n.right, p);
                    if (n.left != null)
                    {
                        double minDist = p.Z - np.Z;
                        if (minDist < smDist)
                            searchSubtreeX(n.left, p);
                    }
                }
            }
        }

        private void addChildren(Node n, List<int> list, int axis, int depth)
        {
            n.axis = axis;

            if (list.Count == 1)
            {
                n.point = list[0];
                return;
            }

            int med = median(list, axis);
            n.point = list[med];

            double medVal = pnts[list[med]][axis];

            List<int> left = new List<int>();
            List<int> right = new List<int>();

            /*for (int i = 0; i < list.Count; i++)
            {
                if (pnts[list[i]][axis] < medVal)
                    left.Add(list[i]);
                else// if (pnts[list[i]][axis] > pnts[list[med]][axis])
                {
                    if (i!=med)
                        right.Add(list[i]);
                }
            }*/

            for (int i = 0; i < list.Count / 2; i++)
                if (i != med)
                {
                    left.Add(list[i]);
                    //if (pnts[list[i]][axis] > medVal)
                    //    Console.WriteLine("Damnit");
                }
            for (int i = list.Count / 2; i < list.Count; i++)
                if (i != med)
                {
                    right.Add(list[i]);
                    //if (pnts[list[i]][axis] < medVal)
                    //    Console.WriteLine("Damnit");
                }


            if (left.Count > 0)
            {
                n.left = new Node();
                addChildren(n.left, left, (axis + 1) % 3, depth + 1);
            }
            if (right.Count > 0)
            {
                n.right = new Node();
                addChildren(n.right, right, (axis + 1) % 3, depth+1);
            }
        }

        public int median(List<int> points, int axis)
        {
            return (findMedian(points, 0, points.Count-1, axis, 0));
        }

        Random rnd = new Random(0);

        private int findMedian(List<int> points, int min, int max, int axis, int depth)
        {
            if (depth > 35)
            {
                //Console.WriteLine("In too deep");
                bool allSame = true;
                for (int i = min + 1; i <= max; i++)
                    if (pnts[points[i]][axis] != pnts[points[min]][axis])
                        allSame = false;
                if (allSame)
                    return (min);
            }
            if (min == max)
                return (min);
            if (min == (max - 1))
            {
                if (pnts[points[min]][axis] > pnts[points[max]][axis])
                {
                    int tmp = points[min];
                    points[min] = points[max];
                    points[max] = tmp;
                }
                return (points.Count / 2);
            }
            int pivot = min + rnd.Next(max - min+1);
            double pivotVal = pnts[points[pivot]][axis];
            int l = min;
            int r = max;
            while (l < r)
            {
                while (pnts[points[l]][axis] < pivotVal)
                    l++;
                while (pnts[points[r]][axis] >= pivotVal)
                {
                    r--;
                    if (r < 0)
                        break;
                }
                if (l < r)
                {
                    int tmp = points[l];
                    points[l] = points[r];
                    points[r] = tmp;
                    //l++;
                    //r--;
                }
            }
            if (r < (points.Count / 2))
            {
                /*for (int i = l; i <= max; i++)
                    if (pnts[points[i]][axis] < pivotVal)
                        Console.WriteLine("Damnit");*/
                return (findMedian(points, l, max, axis, depth+1));
            }
            else
            {
                /*for (int i = min; i <= r; i++)
                    if (pnts[points[i]][axis] > pivotVal)
                        Console.WriteLine("Damnit");*/
                return (findMedian(points, min, r, axis, depth+1));
            }
        }
    }    
}
