﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
using Framework;

namespace MeshBrowser.SlimDXRenderer
{
    public delegate void ProgressInfo(double progress);

    public class Loader
    {        

        public TriangleMesh Load(string fn, ProgressInfo pi)
        {
            try
            {
                //ProgressInfo pi = new ProgressInfo(this.setProgress);

                Directory.SetCurrentDirectory(Application.ExecutablePath.Substring(0, Application.ExecutablePath.LastIndexOf("\\")));



                TriangleMesh mesh = null;

                if (fn.EndsWith(".obj") || fn.EndsWith(".OBJ"))
                {

                    // create an instance of OBJ file parser/loader
                    ObjLoader loader = new ObjLoader();
                    loader.FlipNormals = true;

                    mesh = loader.Execute(fn);
                }
                else if (fn.EndsWith(".off") || fn.EndsWith(".OFF"))
                {

                    // create an instance of OBJ file parser/loader
                    OFFLoader loader = new OFFLoader();

                    mesh = loader.Execute(fn, pi);
                }
                else
                {
                    MeshLoader loader = new MeshLoader();
                    loader.FileName = fn;
                    loader.Progress = pi;
                    loader.Execute();
                    mesh = loader.Mesh;
                }

                if (mesh.Normals == null)
                {
                    //Console.WriteLine("Computing normals");
                    //mesh.initCornerTable();
                    //mesh.ComputeNormals();
                }
                if (mesh.Points.Length <= 0)
                    throw new Exception();
                if (mesh.Triangles.Length <= 0)
                    throw new Exception();
                return (mesh);
            }
            catch(Exception e)
            {
                System.Windows.Forms.MessageBox.Show("Unable to load file " + fn + ", please contact lvasa@kiv.zcu.cz to resolve te issue.");
                GeoidSource gs = new GeoidSource();
                gs.Execute(10, new Point3D(0,0,0));
                return(gs.result);
            }
        }        
    }
}
