﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media.Imaging;
using Framework;
using System.IO;
using SlimDXRenderSystem.Managers;
using System.Threading;
using System.Diagnostics;
using System.Reflection;

namespace MeshBrowser.SlimDXRenderer
{
    /// <summary>
    /// Interaction logic for Window1.xaml
    /// </summary>
    public partial class Window1 : Window
    {
        string openFileName;
        string[] args;
        bool lockCamera = false;
        TriangleMesh[] meshes;
        bool showColorAttribute;
        bool showVectorAttribute;

        public Window1(TriangleMesh[] inputMeshes, bool showColors, bool showVectors)
        {
            meshes = inputMeshes;
            args = Environment.GetCommandLineArgs();
            InitializeComponent();
            Loaded += new RoutedEventHandler(this.onLoaded);
            showColorAttribute = showColors;
            showVectorAttribute = showVectors;
        }

        private void onLoaded(object sender, RoutedEventArgs e)
        {
            initializing = true;
            this.m_slimDXControl.InitManagers(this);


            ModuleManager.Instance.ClientSize = new Size(this.m_slimDXControl.ActualWidth, this.m_slimDXControl.ActualHeight);
            ModuleManager.Instance.Setup(new Framework.ColorRGBA(127, 127, 127, 0), meshes.Length, false, showColorAttribute, showVectorAttribute, false);
            this.propGrid.SelectedObject = ModuleManager.Instance.Rendersetting;
            
            RenderManager.Instance.WizardShown = false;
            //loadMesh(showingWizard);
            loadMeshes(meshes);

            Initialize();
            m_slimDXControl.addRenderingCallback();
            m_slimDXControl.InitManagers2(this);
            SlimDXRenderSystem.Managers.InputManager.Instance.AddKeyClickAction(SlimDX.DirectInput.Key.PageDown, this.LoadNextMesh);
            SlimDXRenderSystem.Managers.InputManager.Instance.AddKeyClickAction(SlimDX.DirectInput.Key.PageUp, this.LoadPreviousMesh);
            SlimDXRenderSystem.Managers.InputManager.Instance.AddControllerSwipeLeftAction(this.LoadNextMesh);
            SlimDXRenderSystem.Managers.InputManager.Instance.AddControllerSwipeRightAction(this.LoadPreviousMesh);
            initializing = false;

            System.Windows.Threading.DispatcherTimer dispatcherTimer = new System.Windows.Threading.DispatcherTimer();
            dispatcherTimer.Tick += new EventHandler(dispatcherTimer_Tick);
            dispatcherTimer.Interval = new TimeSpan(0, 0, 0, 0, 40);
            dispatcherTimer.Start();
        }

        private void dispatcherTimer_Tick(object sender, EventArgs e)
        {
            if (ModuleManager.Instance.Rendersetting.PlayAnimation)
                ModuleManager.Instance.CurrentMesh++;
        }

        private void loadMeshes(TriangleMesh[] meshes)
        {
            GeoidSource gs1 = new GeoidSource();
            gs1.Depth = 1;
            gs1.Execute(1.0, new Point3D(0, 0, 0));
            setFileName("*no file loaded*");

            TriangleMesh input = gs1.result;

            ModuleManager.Instance.SetTransforms(input);

            GeoidSource gs2 = new GeoidSource();
            gs2.Execute(input.BoundingBox.MaxDiag / 200, input.BoundingBox.MidPoint);
            ModuleManager.Instance.SetCursor(gs2.result);
            ModuleManager.Instance.SetMesh(0, input);

            this.m_slimDXControl.Rendering = false;

            new Thread(delegate ()
            {
                while (initializing)
                    System.Threading.Thread.Sleep(100);                

                if (!lockCamera)
                    ModuleManager.Instance.SetTransforms(meshes[0]);

                GeoidSource gs = new GeoidSource();
                gs.Execute(meshes[0].BoundingBox.MaxDiag / 200, meshes[0].BoundingBox.MidPoint);

                SceneManager.Instance.ClearScene();

                ModuleManager.Instance.SetCursor(gs.result);
                for (int i = 0;i<meshes.Length;i++)
                    ModuleManager.Instance.SetMesh(i, meshes[i]);

                SceneManager.Instance.Initialize();

                this.Dispatcher.Invoke(new Action(this.Initialize));
                this.Dispatcher.Invoke(new Action(delegate () { this.m_slimDXControl.Rendering = true; }));
            }).Start();
        }        

        private void showWelcomeWizard()
        {
            RenderManager.Instance.WizardShown = true;
            WelcomeWizard wizard = new WelcomeWizard();
            wizard.ShowOnNextRun.IsChecked = (bool)Properties.Settings.Default["ShowWizard"];
            wizard.ShowDialog();
            Properties.Settings.Default["ShowWizard"] = wizard.ShowOnNextRun.IsChecked;
            Properties.Settings.Default.Save();
            RenderManager.Instance.WizardShown = false;
            //rendererControl1.Invalidate();
        }

        protected override void OnInitialized(EventArgs e)
        {
            base.OnInitialized(e);
        }        

        void loadMesh(bool wizard)
        {
            GeoidSource gs = new GeoidSource();
            gs.Depth = 1;
            gs.Execute(1.0, new Point3D(0, 0, 0));
            setFileName("*no file loaded*");

            TriangleMesh input = gs.result;

            ModuleManager.Instance.SetTransforms(input);

            GeoidSource gs2 = new GeoidSource();
            gs2.Execute(input.BoundingBox.MaxDiag / 200, input.BoundingBox.MidPoint);
            ModuleManager.Instance.SetCursor(gs.result);
            ModuleManager.Instance.SetMesh(0, input);

            if (args.Length < 2)
            {
                var ofd = GetOpenFileDialog();
                if (wizard)
                {
                    // wizard has been shown previously. Try to navigate to the meshes folder
                    string path = System.IO.Path.GetDirectoryName(Environment.GetCommandLineArgs()[0]);
                    path = path.Substring(0, path.LastIndexOf("\\"));
                    ofd.InitialDirectory = path + "\\Meshes";
                }
                if (ofd.ShowDialog() == true)
                {
                    args = new string[] { args[0], ofd.FileName };
                }
                else
                {
                    return;
                }
            }

            loadMesh(args[1]);

        }

        private static Microsoft.Win32.OpenFileDialog GetOpenFileDialog()
        {
            var ofd = new Microsoft.Win32.OpenFileDialog();
            ofd.Filter = "Al supported formats|*.cmx;*.obj;*.off;*.ply;*.stl;*.tri;*.wrl|CMX File|*.cmx|OBJ file|*.obj|OFF file|*.off|PLY file|*.ply|STL file|*.stl|TRI file|*.tri|WRL file|*.wrl|All files|*.*";
            return ofd;
        }

        public void setFileName(string fileName)
        {
            string bit = "(32-bit)";
            if (IntPtr.Size == 8)
                bit = "(64-bit)";
            this.Title = "MeshBrowser"+bit+" - " + fileName;
            this.openFileName = fileName;
        }

        private void loadMesh(string fn)
        {
            this.setFileName(fn);

            Loader loader = new Loader();

            ProgressBar.DataContext = loader;

            this.m_slimDXControl.Rendering = false;

            new Thread(delegate()
            {

                TriangleMesh[] tm = new TriangleMesh[1];
                tm[0] = loader.Load(fn, this.updateProgress);
                while (initializing)
                    System.Threading.Thread.Sleep(100);
                TriangleMesh input = null;

                input = tm[0];

                if (!lockCamera)
                    ModuleManager.Instance.SetTransforms(input);

                GeoidSource gs = new GeoidSource();
                gs.Execute(input.BoundingBox.MaxDiag / 200, input.BoundingBox.MidPoint);

                SceneManager.Instance.ClearScene();

                ModuleManager.Instance.SetCursor(gs.result);
                ModuleManager.Instance.SetMesh(0, input);
                
                SceneManager.Instance.Initialize();
                
                this.Dispatcher.Invoke(new Action(this.Initialize));
                this.Dispatcher.Invoke(new Action(delegate() { this.m_slimDXControl.Rendering = true; }));
            }).Start();
            

            //SceneManager.Instance.Initialize();
            //SceneManager.Instance.ClearScene();
            //Initialize();
        }        

        private void updateProgress(double progress)
        {
            this.Dispatcher.Invoke(System.Windows.Threading.DispatcherPriority.Render, new Action(delegate() { ProgressBar.Value = progress * 100; })); 
        }

        bool initializing = false;

        public void Initialize()
        {
            //initializing = true;
            lock(this)
            {
                ModuleManager.Instance.Initialize(lockCamera);

                //Scene Tree view setting
                this.treeSceneView.Items.Clear();

                TreeViewItem tn = new TreeViewItem() { Header = "Meshes" };
                treeSceneView.Items.Add(tn);
                foreach (SlimDXRenderSystem.Core.Group grp in SceneManager.Instance.SceneGroups.Values)
                {
                    if (grp.NumberOfObjects > 0)
                    {
                        TreeViewItem gtn = new TreeViewItem() { Header = grp.Name };
                        tn.Items.Add(gtn);
                        gtn.Tag = grp;
                        foreach (SlimDXRenderSystem.Core.Object3D obj in grp.Objects.Values)
                        {
                            TreeViewItem obtn = new TreeViewItem() { Header = obj.Name };
                            gtn.Items.Add(obtn);
                            obtn.Tag = obj;
                            obtn.Items.Add(new TreeViewItem() { Header = "Material", Tag = obj.Material });
                        }
                    }
                }

                tn = new TreeViewItem() { Header = "Cameras" };
                this.treeSceneView.Items.Add(tn);
                foreach (SlimDXRenderSystem.Core.Camera cam in SceneManager.Instance.SceneCameras.Values)
                {
                    tn.Items.Add(new TreeViewItem() { Header = cam.Name, Tag = cam });
                }

                tn = new TreeViewItem() { Header = "Lights" };
                this.treeSceneView.Items.Add(tn);
                foreach (SlimDXRenderSystem.Core.Light light in SceneManager.Instance.SceneLights.Values)
                {
                    tn.Items.Add(new TreeViewItem() { Header = light.Name, Tag = light });
                }

                ModuleManager.Instance.VertexPickAction += this.VertexPick;
                ModuleManager.Instance.PrimitivePickAction += this.PrimitivePick;
            }
            //initializing = false;
        }

        private void PrimitivePick(int index, int v1, int v2, int v3)
        {
            this.PickedCoordinatesLabel.Content = "Triangle " + index.ToString() + " (" + v1.ToString() + ", " + v2.ToString() + ", " + v3.ToString() + ") selected";
        }

        private void VertexPick(int index, double x, double y, double z)
        {
            this.PickedCoordinatesLabel.Content = "Vertex " + index.ToString() + " (x: " + x.ToString() + ", y: " + y.ToString() + ", z: " + z.ToString() + ") selected";
        }

        private void MenuItem_Click(object sender, RoutedEventArgs e)
        {
            this.Close();
        }

        private void MenuItem_Click_1(object sender, RoutedEventArgs e)
        {
            var ofd = GetOpenFileDialog();
            if (ofd.ShowDialog() == true)
            {
                loadMesh(ofd.FileName);
            }
        }

        private void Button_Click(object sender, RoutedEventArgs e)
        {
            ModuleManager.Instance.Rendersetting.SaveCameraPosition();
            this.CameraLoadButton.IsEnabled = true;
        }

        private void CameraLoadButton_Click(object sender, RoutedEventArgs e)
        {
            ModuleManager.Instance.Rendersetting.LoadCameraPosition();
        }

        private void Button_Click_1(object sender, RoutedEventArgs e)
        {
            ModuleManager.Instance.Rendersetting.SaveSingleFrame();
        }

        private void Button_Click_2(object sender, RoutedEventArgs e)
        {
            LoadPreviousMesh();
        }

        private void LoadNextMesh()
        {
            if (openFileName.StartsWith("*"))
            {
                loadMesh(false);
                
            }
            else
            {
                string[] files = Directory.GetFiles(openFileName.Substring(0, this.openFileName.LastIndexOf("\\")));
                List<string> supported = new List<string>();
                foreach (string s in files)
                {
                    if (isFileSupported(s))
                        supported.Add(s);
                }
                for (int i = 0; i < supported.Count; i++)
                {
                    if (supported[i] == openFileName)
                    {
                        if ((i + 1) < supported.Count)
                            loadMesh(supported[i + 1]);
                        else
                            loadMesh(supported[0]);

                        /*SceneManager.Instance.Initialize();
                        Initialize();*/

                        break;
                    }
                }
            }
        }

        private void LoadPreviousMesh()
        {
            if (openFileName.StartsWith("*"))
            {
                loadMesh(false);
                /*SceneManager.Instance.Initialize();
                Initialize();*/
            }
            else
            {
                string[] files = Directory.GetFiles(openFileName.Substring(0, this.openFileName.LastIndexOf("\\")));
                List<string> supported = new List<string>();
                foreach (string s in files)
                {
                    if (isFileSupported(s))
                        supported.Add(s);
                }
                for (int i = 0; i < supported.Count; i++)
                {
                    if (supported[i] == openFileName)
                    {
                        if ((i - 1) < 0)
                            loadMesh(supported[supported.Count - 1]);
                        else
                            loadMesh(supported[i - 1]);

                        /*SceneManager.Instance.Initialize();
                        Initialize();*/

                        break;
                    }
                }
            }
        }

        bool isFileSupported(string fn)
        {
            if (fn.EndsWith(".obj"))
                return (true);
            if (fn.EndsWith(".OBJ"))
                return (true);
            if (fn.EndsWith(".wrl"))
                return (true);
            if (fn.EndsWith(".WRL"))
                return (true);
            if (fn.EndsWith(".tri"))
                return (true);
            if (fn.EndsWith(".TRI"))
                return (true);
            if (fn.EndsWith(".stl"))
                return (true);
            if (fn.EndsWith(".STL"))
                return (true);
            if (fn.EndsWith(".cmx"))
                return (true);
            if (fn.EndsWith(".CMX"))
                return (true);
            if (fn.EndsWith(".ply"))
                return (true);
            if (fn.EndsWith(".PLY"))
                return (true);
            if (fn.EndsWith(".off"))
                return (true);
            if (fn.EndsWith(".OFF"))
                return (true);


            return (false);
        }

        private void Button_Click_3(object sender, RoutedEventArgs e)
        {
            LoadNextMesh();
        }

        private void Button_Click_4(object sender, RoutedEventArgs e)
        {
            ModuleManager.Instance.Recenter();
        }

        private void Button_Click_5(object sender, RoutedEventArgs e)
        {
            ModuleManager.Instance.Rendersetting.ToggleShowNormals();
        }

        private void Button_Click_6(object sender, RoutedEventArgs e)
        {
            ModuleManager.Instance.Rendersetting.ToggleLighting();
        }

        private void Button_Click_7(object sender, RoutedEventArgs e)
        {
            lockCamera = !lockCamera;
            CameraLockButton.IsChecked = lockCamera;
        }

        private void ToggleButton_Click(object sender, RoutedEventArgs e)
        {
            ModuleManager.Instance.Pick = PickType.NONE;
            pickPoints.IsChecked = false;
            pickPrimitives.IsChecked = false;
            manipulate.IsChecked = true;
        }

        private void pickPoints_Click(object sender, RoutedEventArgs e)
        {
            ModuleManager.Instance.Pick = PickType.POINTS;
            pickPoints.IsChecked = true;
            pickPrimitives.IsChecked = false;
            manipulate.IsChecked = false;
        }

        private void pickPrimitives_Click(object sender, RoutedEventArgs e)
        {
            ModuleManager.Instance.Pick = PickType.PRIMITIVES;
            pickPoints.IsChecked = false;
            pickPrimitives.IsChecked = true;
            manipulate.IsChecked = false;
        }

        private void Button_Click_8(object sender, RoutedEventArgs e)
        {
            this.propGrid.SelectedObject = ModuleManager.Instance.Rendersetting;
        }

        private void treeSceneView_SelectedItemChanged(object sender, RoutedPropertyChangedEventArgs<object> e)
        {
            if (treeSceneView.SelectedItem!=null)
                this.propGrid.SelectedObject = (treeSceneView.SelectedItem as TreeViewItem).Tag;
        }

        private void m_slimDXControl_MouseMove(object sender, MouseEventArgs e)
        {
            ModuleManager.Instance.MouseX = e.GetPosition(m_slimDXControl).X;
            ModuleManager.Instance.MouseY = e.GetPosition(m_slimDXControl).Y;
        }

        private void Button_Click_10(object sender, RoutedEventArgs e)
        {
            FindVertex fvd = new FindVertex();
            fvd.IndexUpDown.Maximum = ModuleManager.Instance.CurrentObject.Points.Length - 1;
            fvd.IndexUpDown.Value = 0;
            if (fvd.ShowDialog() == true)
            {
                this.pickPoints_Click(sender, e);
                ModuleManager.Instance.PickVertex((int)fvd.IndexUpDown.Value);
            }
        }

        private void Button_Click_9(object sender, RoutedEventArgs e)
        {
            FindVertex fvd = new FindVertex();
            fvd.Title = "Find Triangle";
            fvd.IndexLabel.Content = "Triangle index:";
            fvd.IndexUpDown.Maximum = ModuleManager.Instance.CurrentObject.Points.Length - 1;
            fvd.IndexUpDown.Value = 0;
            if (fvd.ShowDialog() == true)
            {
                this.pickPrimitives_Click(sender, e);
                ModuleManager.Instance.PickPrimitive((int)fvd.IndexUpDown.Value);
            }
        }

        private void MenuItem_Click_2(object sender, RoutedEventArgs e)
        {
            AboutDialog about = new AboutDialog();
            about.ShowDialog();
        }

        private void MenuItem_Click_3(object sender, RoutedEventArgs e)
        {
            showWelcomeWizard();
        }

        private void MenuItem_Click_4(object sender, RoutedEventArgs e)
        {
            AssociationsDialog form = new AssociationsDialog();
            if (form.ShowDialog() == true)
            {
                List<string> toAdd = new List<string>();
                if ((bool)form.cbObj.IsChecked && (!form.orig[0])) toAdd.Add(".obj");
                if ((bool)form.cbTri.IsChecked && (!form.orig[1])) toAdd.Add(".tri");
                if ((bool)form.cbStl.IsChecked && (!form.orig[2])) toAdd.Add(".stl");
                if ((bool)form.cbWrl.IsChecked && (!form.orig[3])) toAdd.Add(".wrl");
                if ((bool)form.cbPly.IsChecked && (!form.orig[4])) toAdd.Add(".ply");
                if ((bool)form.cbCmx.IsChecked && (!form.orig[5])) toAdd.Add(".cmx");
                if ((bool)form.cbCmx.IsChecked && (!form.orig[6])) toAdd.Add(".off");

                if (toAdd.Count > 0)
                {
                    var psi = new ProcessStartInfo();
                    psi.FileName = Environment.GetCommandLineArgs()[0];
                    psi.Arguments = "-a ";
                    foreach (string s in toAdd)
                        psi.Arguments += " " + s;

                    psi.Verb = "runas";

                    var process = new Process();
                    process.StartInfo = psi;
                    process.Start();
                    process.WaitForExit();

                }
            }
        }

        private void MenuItem_Click_5(object sender, RoutedEventArgs e)
        {
            ControlsWindow csw = new ControlsWindow();
            csw.ShowDialog();
        }

        private void MenuItem_Click_6(object sender, RoutedEventArgs e)
        {
            (new SettingsWindow()).ShowDialog(); 
        }
    }
}
