﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.ComponentModel;
using Framework;

namespace MeshBrowser.SlimDXRenderer
{
    public class OFFLoader
    {
        public TriangleMesh _res;

        private string _filename = "";

        private bool _reload = false;

        public string FileName
        {
            get { return _filename; }
            set { _filename = value; }
        }

        public bool Reload
        {
            get { return _reload; }
            set { _reload = value; }
        }

        public OFFLoader()
        {
        }

        private string ReadLine(StreamReader sr)
        {
            string res;

            do
            {
                res = sr.ReadLine().Trim();
            }
            while (res.Length == 0 || res[0] == '#');

            return res;
        }

        private void LoadFile(string fileName, ProgressInfo progress)
        {
            _filename = fileName;

            if (!File.Exists(_filename))
            {
                throw new Exception("File " + _filename + " does not exist.");
            }

            StreamReader sr;
            string line;
            string[] tokens;

            int pCount, fCount, fSize;

            System.Globalization.CultureInfo ci = System.Globalization.CultureInfo.GetCultureInfo("en-US");

            sr = new StreamReader(_filename);

            line = ReadLine(sr);

            if (!line.Equals("OFF")) throw new Exception("Not an OFF file.");

            line = ReadLine(sr);
            tokens = line.Split(' ');

            if (tokens.Length < 2) throw new Exception("Wrong file format.");

            pCount = int.Parse(tokens[0].Trim(), ci);
            fCount = int.Parse(tokens[1].Trim(), ci);
            double frac = 1.0 / (pCount + fCount);
            int pStep = pCount / 10;
            int fStep = fCount / 10;

            _res = new TriangleMesh();
            Point3D[] pts = new Point3D[pCount];
            Triangle[] cells = new Triangle[fCount];
            
            for (int i = 0, pi = pStep; i < pCount; i++, pi++)
            {
                line = ReadLine(sr);
                tokens = line.Split(' ');

                if (tokens.Length < 3) throw new Exception("Wrong file format.");

                pts[i] = new Point3D(
                    double.Parse(tokens[0].Trim(), ci),
                    double.Parse(tokens[1].Trim(), ci),
                    double.Parse(tokens[2].Trim(), ci));
                if (i % 10000 == 0)
                    progress(((double)i / pCount) / 2);
            }

            for (int i = 0, fi = fStep; i < fCount; i++, fi++)
            {
                line = ReadLine(sr);
                tokens = line.Split(new char[]{' '}, StringSplitOptions.RemoveEmptyEntries);
                fSize = int.Parse(tokens[0], ci);

                if (fSize != 3) throw new Exception("Cells other than triangles are currently not supported.");
                if (tokens.Length < 4) throw new Exception("Wrong file format.");

                cells[i] = new Triangle(
                    int.Parse(tokens[1].Trim(), ci),
                    int.Parse(tokens[2].Trim(), ci),
                    int.Parse(tokens[3].Trim(), ci));
                if (i % 10000 == 0)
                    progress(0.5+((double)i / fCount) / 2);
            }

            _res.Points = pts;
            _res.Triangles = cells;
        }

        public TriangleMesh Execute(string fn, ProgressInfo pi)
        {
            LoadFile(fn, pi);
            return (_res);
        }
    }
}
