using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

using Framework;
using System.Globalization;

namespace MeshBrowser.SlimDXRenderer
{
	class ReaderWrl
	{
		public DataSet ReadFile(string fileName)
		{
			StreamReader sr = new StreamReader(fileName);
			int pointCount = 0;
			int triangleCount = 0;
			while (!sr.EndOfStream)
			{
				string line = sr.ReadLine();
				if (line.Contains("point ["))
				{
					line = sr.ReadLine();
                    int coordCount = 0;
					while (!line.Contains("]"))
					{
                        coordCount += line.Split(new char[] { ' ', ',', }, StringSplitOptions.RemoveEmptyEntries).Length;
						line = sr.ReadLine();
					}
                    string substr = line.Substring(0, line.IndexOf(']'));
                    coordCount += substr.Split(new char[] { ' ', ',', }, StringSplitOptions.RemoveEmptyEntries).Length;

                    pointCount = coordCount/3;
				}

				if (line.Contains("coordIndex ["))
				{
                    line = sr.ReadLine();
                    int indexCount = 0;
                    while (!line.Contains("]"))
                    {
                        indexCount += line.Split(new char[] { ' ', ',', }, StringSplitOptions.RemoveEmptyEntries).Length;
                        line = sr.ReadLine();
                    }
                    string substr = line.Substring(0, line.IndexOf(']'));
                    indexCount += substr.Split(new char[] { ' ', ',', }, StringSplitOptions.RemoveEmptyEntries).Length;

                    triangleCount = indexCount / 4; // this is wrong, maybe there are polygons
				}
			}

			DataSet result = new DataSet(pointCount, triangleCount, 0);

			sr.Close();
			sr = new StreamReader(fileName);
			int pointIndex = 0;
			int triangleIndex = 0;
            NumberFormatInfo nfi = new NumberFormatInfo();
            nfi.NumberDecimalSeparator = ".";
			while (!sr.EndOfStream)
			{
				string line = sr.ReadLine();
				if (line.Contains("point ["))
				{
					line = sr.ReadLine();
                    List<double> coordinates = new List<double>();
                    string[] parts;
					while (!line.Contains("]"))
					{
						parts = line.Split(new char[] { ' ', ',' }, StringSplitOptions.RemoveEmptyEntries);
                        foreach (string s in parts)
                            coordinates.Add(double.Parse(s, nfi));
						
						
						line = sr.ReadLine();
					}
                    string substr = line.Substring(0, line.IndexOf(']'));
                    parts = substr.Split(new char[] { ' ', ',' }, StringSplitOptions.RemoveEmptyEntries);
                    foreach (string s in parts)
                        coordinates.Add(double.Parse(s, nfi));


                    for (int i = 0;i<coordinates.Count/3;i++)
                    {
                        result.AddVertex(pointIndex, new Point3D(coordinates[3*i], coordinates[3*i+1], -coordinates[3*i+2]));
                        pointIndex++;
                    }
				}

				if (line.Contains("coordIndex ["))
				{
					line = sr.ReadLine();
                    List<int> indices = new List<int>();
                    string[] parts;
					while (!line.Contains("]"))
					{
						parts = line.Split(new char[] { ' ', ',' }, StringSplitOptions.RemoveEmptyEntries);
                        foreach (string s in parts)
                            indices.Add(int.Parse(s));

						line = sr.ReadLine();						
					}
                    string substr = line.Substring(0, line.IndexOf(']'));
                    parts = substr.Split(new char[] { ' ', ',' }, StringSplitOptions.RemoveEmptyEntries);
                    foreach (string s in parts)
                        indices.Add(int.Parse(s));

                    for(int i = 0;i<indices.Count/4;i++)
                    {
                        result.AddTriangle(triangleIndex, new Triangle(indices[4*i], indices[4*i+2], indices[4*i+1]));
                        triangleIndex++;
                        if (indices[4 * i + 3] != -1)
                            throw new Exception("General polygon (non-triangle) found in WRL. Cannot continue.");
                    }
				}
			}
			return (result);
		}
	}
}
